/*
 * Copyright (c) 2004-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 180 $
 * $LastChangedDate: 2007-06-20 11:22:10 +0200 (Wed, 20 Jun 2007) $
 * 
 */

package com.codedroids.j;

import java.util.*;
import java.io.*;

public class CSVReader extends Object implements Iterator {
	
	public static final String NullValue = "<NULL>";
	private BufferedReader reader;
	private int recordCount;
	private String recordSeparator; // currently ignored
	private String fieldSeparator;
	private String quoteCharacter;
	private String encoding;
	private boolean firstRecordContainsLabels;
	private List labels;
	private Object nullValue;
	private boolean patchMissingColumns = true;
	
	/**
	 * Creates a reader with the default separators, " as quote character and
	 * assuming the first line holds labels
	 *
	 */
	public CSVReader() 
	{
		this.setRecordSeparator(null);
		this.setFieldSeparator(null);
		this.setQuoteCharacter("\"");
		this.setFirstRecordContainsLabels(true);
		this.setNullValue(null);
	}
	
	/**
	 * Creates a reader with the given properties. If null are passed as value to a
	 * parameter, this means use the default parameter (see below).
	 * 
	 * @param recSep
	 * @param fieldSep
	 * @param quoteChar
	 * @param labels
	 * @param nullValue
	 */
	public CSVReader(String recSep, String fieldSep, String quoteChar, boolean labels, Object nullValue) 
	{
		this.setRecordSeparator(recSep);
		this.setFieldSeparator(fieldSep);
		this.setQuoteCharacter(quoteChar);
		this.setFirstRecordContainsLabels(labels);
		this.setNullValue(nullValue);
	}
	
	/**
	 * Creates a reader with the given properties. If null are passed as value to a
	 * parameter, this means use the default parameter (see below).
	 * 
	 * @param recSep
	 * @param fieldSep
	 * @param quoteChar
	 * @param enc
	 * @param labels
	 * @param nullValue
	 * @deprecated Use the variant without the encoding, and then set the encoding directly when invoking <code>open()</code>
	 */
	public CSVReader(String recSep, String fieldSep, String quoteChar, String enc, boolean labels, Object nullValue) 
	{
		this.setRecordSeparator(recSep);
		this.setFieldSeparator(fieldSep);
		this.setQuoteCharacter(quoteChar);
		this.setEncoding(enc);
		this.setFirstRecordContainsLabels(labels);
		this.setNullValue(nullValue);
	}
	
	
	protected void finalize() throws Throwable
	{
		if(reader != null) {
			reader.close();
		}
		super.finalize();
	}
	
	/**
	 * Sets the string that separates the records. 
	 * 
	 * @param recSep If null the Java System property "line.separator" is used
	 */
	public void setRecordSeparator(String recSep)
	{
		if(recSep != null)
			recordSeparator = recSep;
		else
			recordSeparator = java.lang.System.getProperty("line.separator");
	}
	
	/**
	 * Returns the record separator.
	 * 
	 * @return
	 */
	public String getRecordSeparator()
	{
		return recordSeparator;
	}
	
	/**
	 * Returns the record separator.
	 * 
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public String recordSeparator()
	{
		return recordSeparator;
	}
	
	/**
	 * Sets the string that separates the fields with a record.
	 * 
	 * @param fieldSep If null it defaults to , (comma)
	 */
	public void setFieldSeparator(String fieldSep)
	{
		if(fieldSep != null)
			fieldSeparator = fieldSep;
		else
			fieldSeparator = ",";
	}
	
	/**
	 * Returns the field separator.
	 * 
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public String getFieldSeparator()
	{
		return fieldSeparator;
	}
	
	/**
	 * Returns the field separator.
	 * 
	 * @return
	 */
	public String fieldSeparator()
	{
		return fieldSeparator;
	}
	
	/**
	 * Sets the quote character.
	 * <p>The quote character is used to escape otherwise significant characters. Typically text fields are
	 * quoted to prevent say a comma to be interpreted as a field separator.
	 * <br>Note some programs does not quote text fields when exporting to CSV, set <code>quoteChar</code> to null
	 * in those cases. 
	 * 
	 * @param quoteChar
	 */
	public void setQuoteCharacter(String quoteChar)
	{
		if(quoteChar != null)
			quoteCharacter = quoteChar;
		else
			quoteCharacter = null;
	}
	
	/**
	 * Returns the character used to qoute text fields.
	 * 
	 * @return
	 */
	public String getQuoteCharacter()
	{
		return quoteCharacter;
	}
	
	/**
	 * Returns the character used to qoute text fields.
	 * 
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public String quoteCharacter()
	{
		return quoteCharacter;
	}
	
	/**
	 * Sets the character encoding for the file to be read.
	 * 
	 * @param enc Setting this to null will cause the file to be read using the systems default encoding
	 * @deprecated Pass the encoding directly to the <code>open</code> method - see below.
	 */
	public void setEncoding(String enc)
	{
		encoding = enc;
	}
	
	/**
	 * File character encoding
	 * @return
	 * @deprecated Pass the encoding directly to the <code>open</code> method - see below.
	 */
	public String encoding()
	{
		return encoding;
	}
	
	/**
	 * If the first record holds the labels for the fields, set this to true.
	 * 
	 * @param flag
	 */
	public void setFirstRecordContainsLabels(boolean flag)
	{
		firstRecordContainsLabels = flag;
	}
	
	/**
	 * Indicates whether or not the first record is treated as field labels
	 * 
	 * @return
	 */
	public boolean getFirstRecordContainsLabels()
	{
		return firstRecordContainsLabels;
	}
	
	/**
	 * Indicates whether or not the first record is treated as field labels
	 * 
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public boolean firstRecordContainsLabels()
	{
		return firstRecordContainsLabels;
	}
	
	/**
	 * Rather than setting nulls into the result, the given null value object is used as null indicator.
	 * 
	 * @param nullVal If set to null, the default <code>CVSReader.NullValue</code> constant is used
	 */
	public void setNullValue(Object nullVal)
	{
		if(nullVal != null)
			nullValue = nullVal;
		else
			nullValue = CSVReader.NullValue; // Same default as in the writer
	}
	
	/**
	 * Returns the object that acts as a null value indicator
	 * @return
	 */
	public Object getNullValue()
	{
		return nullValue;
	}
	/**
	 * Returns the object that acts as a null value indicator
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public Object nullValue()
	{
		return nullValue;
	}
	
	/**
	 * Convenience that will return the standin null object if passed value is null
	 * @param o
	 * @return
	 */
	public Object ifNull(Object o) {
		if(o==null)
			return getNullValue();
		else
			return o;
	}
	
	/**
	 * If the first record does not hold the field labels, they must be provided if record maps are needed.
	 * 
	 * @param arr A list of strings corresponding to the fields in the data set
	 */
	public void setLabels(List arr)
	{
		labels = arr;
	}
	
	/**
	 * Returns the list of field labels
	 * @return
	 */
	public List getLabels()
	{
		return labels;
	}
	
	/**
	 * 
	 * @return
	 * @deprecated Use the get... ditto - this is a WebObjects/Objective C naming convention relic
	 */
	public List labels()
	{
		return labels;
	}

	//
	// Set data to be read
	//
	
	/**
	 * Set a reader from which the data should be read.
     * <p>If the <code>aReader</code> is not already buffered it will be.
     * 
	 * @param aReader
	 * @throws IOException
	 */
	public void open(Reader aReader) throws IOException
	{
		if(reader != null)
			reader.close();
		if(aReader instanceof BufferedReader)
			reader = (BufferedReader)aReader;
		else
			reader = new BufferedReader(aReader);
		recordCount = 0;
	}
	
	/**
	 * Opens reader on inputstram with the given encoding.
	 * 
	 * @param is
	 * @param enc If null the default system encoding is used
	 */
	public void open(InputStream is, String enc) throws IOException
	{
		open(new InputStreamReader(is, enc));
	}
	
	/**
	 * Opens reader on byte array with the given encoding.
	 * 
	 * @param bytes
	 * @param enc If null the default system encoding is used
	 */
	public void open(byte[] bytes, String enc) throws IOException
	{
		open(new InputStreamReader(new ByteArrayInputStream(bytes), enc));
	}
	
	/**
	 * Opens a file with the system default encoding
	 * <p>or the encoding set previously (note that this 'extra' behaviour is deprecated) 
	 * @param aFileName
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public void open(String aFileName) throws FileNotFoundException, IOException
	{
		if(encoding==null)
			open(new FileReader(aFileName));
		else
			open(new InputStreamReader(new FileInputStream(aFileName), encoding));
	}
	
	/**
	 * Opens a file with the given encoding.
	 * 
	 * @param aFileName
	 * @param enc If null the default system encoding is used
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public void open(String aFileName, String enc) throws FileNotFoundException, IOException
	{
		if(enc==null)
			open(new FileReader(aFileName));
		else
			open(new InputStreamReader(new FileInputStream(aFileName), enc));
	}
	
	/**
	 * Closes the Reader used here
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException
	{
		if(reader != null) {
			reader.close();
		}
	}
	
	private String _prefetchedRecord;
	
	/**
	 * Returns the next record from the data set.
	 * <p>Throws in case no more records are available.
	 * @return A string representing one record from the data et
	 * @throws NoSuchElementException
	 * @throws IOException
	 */
	public String nextRecord() throws NoSuchElementException, IOException
	{
		String retval = null;
		if(_prefetchedRecord == null) {
			if(reader != null) {
				retval = reader.readLine();
				recordCount++;
			}
			if(retval == null) {
				reader.close();
				reader = null;
				throw new NoSuchElementException("No more records");
			}
		} else {
			retval = _prefetchedRecord;
			_prefetchedRecord = null;
		}
		return retval;
	}
	
	/**
	 * Tests whether or not the data set contains more records
	 * @return
	 * @throws IOException
	 */
	public boolean hasMoreRecords() throws IOException {
		if(_prefetchedRecord == null) {
			if(reader != null) {
				_prefetchedRecord = reader.readLine();
				recordCount++;
			}
			if(_prefetchedRecord == null) {
				reader.close();
				reader = null;
			}
		}
		return (_prefetchedRecord != null);
	}

	/**
	 * Returns the next record from the data set.
	 * <p>Implements the <code>Iterator</code> interface.
	 * <p>Note it will throw a <code>NoSuchElementException</code> in case of IO errors
	 * @return Either a <code>Map</code> (in case the fields labels are known) or else a <code>List</code> object
	 * @throws NoSuchElementException
	 */
	public Object next() throws NoSuchElementException
	{
		try {
			// If not the label stuff seems to be in place we use arrays.
			if(firstRecordContainsLabels || (labels != null))
				return nextRecordMap();
			else
				return nextRecordList();
		} catch (IOException e) {
			throw new NoSuchElementException("IOException: " + e.getMessage());
		}
	}
	
	/**
	 * Checks if there are more records in the data set.
	 * <p>Implements the <code>Iterator</code> interface.
	 * <p>Note it will return <code>false</code> in case of IO errors
	 * @return
	 */
	public boolean hasNext() 
	{
		try {
			return hasMoreRecords();
		} catch (IOException e) {
			return false; // What else to do...
		}
	}
	
	/**
	 * No op - just here to satisfy the <code>Iterator</code> interface
	 */
	public void remove() {
		// since the interface defines it (albeit as optional)... 
	}
	
	/**
	 * Returns the next record as a list
	 * @return
	 * @throws NoSuchElementException
	 * @throws IOException
	 */
	public List nextRecordList() throws NoSuchElementException, IOException
	{
		String rec = nextRecord();
		CSVRecordReader rr = new CSVRecordReader(rec);
		ArrayList arr = new ArrayList();
		Object tmp;
		while((tmp = rr.nextField()) != null) {
			arr.add(tmp);
		}
		return arr;
	}
	
	/**
	 * Returns the next record as a Map - keys in the map are the field labels
	 * <p>Throws in case the no labels are found, or the number of labels does not match the number of fields
	 * @return
	 * @throws NoSuchElementException
	 * @throws IOException
	 */
	public Map nextRecordMap() throws NoSuchElementException, IOException
	{
		List arr = nextRecordList();
		if(firstRecordContainsLabels && recordCount==1) {
			labels = arr;
			arr = nextRecordList();
		}
		
		// A minor hack as some exporters do not fill in 'blanks' in the end of the list.
		while(patchMissingColumns && labels.size() > arr.size()) {
			arr.add(getNullValue());
		}
		if(labels.size() == arr.size())
			return MapUtils.getHashMap(labels, arr);
		else
			throw new NoSuchElementException("Mismatch in number of labels: "+labels.toString()+" and values: "+arr.toString());
	}
	
	
	private class CSVRecordReader extends Object
	{
		private String source;
		private int currentPosition;
		private int maxPosition;
		
		CSVRecordReader(String line) {
			source = line;
			currentPosition = 0;
			maxPosition = line.length();
		}
		
		private int nextFieldSeparator(int ind) {
			boolean inquote = false;
			while (ind < maxPosition) {
				char ch = source.charAt(ind);
				if (!inquote && ch == fieldSeparator.charAt(0)) {
					break;
				} else if (quoteCharacter != null && ch == quoteCharacter.charAt(0)) {
					inquote = !inquote;
				}
				ind ++;
			}
			return ind;
		}
		
		int countTokens() {
			int i = 0;
			int ret = 1;
			while ((i = nextFieldSeparator(i)) < maxPosition) {
				i ++;
				ret ++;
			}
			return ret;
		}
		
		
		Object nextField() {
			if (currentPosition > maxPosition)
				return null;
			
			boolean inquote = false;
			int st = currentPosition;
			currentPosition = nextFieldSeparator(currentPosition);
			StringBuffer buf = null;
			
			if(nullValue == null)
				buf = new StringBuffer(); // no!? then we at least returns an empty string
			while (st < currentPosition) {
				char ch = source.charAt(st++);
				if (quoteCharacter != null && ch == quoteCharacter.charAt(0)) {
					if(inquote) {
						if(buf==null) 
							buf = new StringBuffer(); // it's at least an empty string then
						if ((st < currentPosition) && (source.charAt(st) == quoteCharacter.charAt(0))) {
							buf.append(ch);
							st ++;
						} else {
							inquote = false;
						}
					} else {
						inquote = true;
					}
				} else {
					if(buf==null) 
						buf = new StringBuffer();
					buf.append(ch);
				}
			}
			// if inquote==true we should complain...
			currentPosition ++;
			return (buf==null ? nullValue : new String(buf));
		}
		
		boolean hasMoreFields() {
			return (nextFieldSeparator(currentPosition) <= maxPosition);
		}
	}
}

