/*
 * Copyright (c) 2005-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 117 $
 * $LastChangedDate: 2007-02-20 11:24:48 +0100 (Tue, 20 Feb 2007) $
 * 
 */

package com.codedroids.jam;

import java.util.List;
import java.util.Stack;
import java.util.Map;
import java.util.HashMap;

/* code example for jsp:

JspApp app = OCJspApp.getInstance(cms.getCmsObject(), "/testarea/test.app.xml");
String prefix = app.getPrefix();

String jamId = request.getParameter(prefix+"id");
Hashtable allInstances;
boolean newInstance = false;
JspAppMachine jam = null;
synchronized(session) {
	allInstances = (Hashtable)session.getAttribute(prefix+name);
	if(allInstances == null) {
		// First time this app is instantiated in this session:
		allInstances = new Hashtable();
		allInstances.put("instanceCount", "0");
	}
	synchronized(allInstances) {
		if(jamId == null || (jam = (JspAppMachine)allInstances.get(jamId)) == null) {
			// Need a new instance - note that if instanceId 
			// was given but it was unknown a new id is generated
			newInstance = true;
			String instanceCountString = (String) allInstances.get("instanceCount");
			int instanceCount = Integer.decode(instanceCountString).intValue()+1;
			// Create a new instance of this machine
			jam = new JspAppMachine(app,""+instanceCount); // count = instance id 
			allInstances.put("instanceCount", ""+instanceCount);
			allInstances.put(jam.getId(), jam);
			session.setAttribute(prefix+name, allInstances);
		}
	}
}

request.setAttribute(prefix+"name", jam.getName()); 
request.setAttribute(prefix+"id", jam.getId());

if(newInstance) {
	// it's a new instance just display the page then
	response.sendRedirect(request.getRequestURI()+"?"+prefix+"id="+jam.getId());
	return;
}

// Perform input action of the current state (if any), this will set the condition
// which in turn defines which transition to choose.
String inputAction = jam.getProcess();
if(!inputAction.equals("")) 
	pageContext.include(inputAction);

// Get condition from request
String condition = (String) request.getAttribute(prefix+"condition");
// If no valid condition, simply redisplay the current page
if(!jam.checkCondition(condition)) {
	pageContext.include(jam.getDisplay());
	return;
}

// Change state now
synchronized(session) {
	allInstances = (Hashtable)session.getAttribute(prefix+name);
	synchronized(allInstances) {
		jam = (JspAppMachine)allInstances.get(jamId);
		synchronized(jam) {
			jam.changeState(condition);
		}
		allInstances.put(jamId, jam);
	}
	session.setAttribute(prefix+name, allInstances);
}
// Display output
pageContext.include(jam.getDisplay());

*/

/**
 * <p>Finite-state machine modelled after the idea presented in
 * http://www.onjava.com/pub/a/onjava/2001/06/27/java_security.html
 * 
 * <p>Builds a finite-state machine from an xml file or programtic which
 * then becomes the is run in a JSP page which guides the user through
 * the sub-pages defined by the state-machine.
 * 
 * <p>These sub-pages must be protected by checking that they are really
 * a part of a machine setup. This way evil doers cannot access them 
 * outside the machine context
 *  
 */
public class JspAppMachine {
	
	//
	// Instance
	//

	private String instanceId;
	private String initState;
	private String[] fsmStates;
	private String[] displayActions;
	private String[] processActions;
	private String[][] inputConditions;
	private int[] numTransitions;
	private int[] cumulativeTransitions;
	private String[] transitionTypes;
	private String[] nextStates;
	private Stack stateStack = new Stack();
	private String currentState;
	private HashMap context = new HashMap();
	
	private int getStateIndex(String state) {
		for(int i=0; i<nextStates.length; ++i) {
			if(fsmStates[i].equals(state))
				return i;
		}
		return -1;
	}
	private int getConditionIndex(int stateIndex, String condition) {
	    if(stateIndex == -1) 
			return -1;
	    for(int i=0; i<inputConditions[stateIndex].length; ++i) {
	        if(inputConditions[stateIndex][i].equals(condition)) {
	            return (cumulativeTransitions[stateIndex] + i);
	        }
	    }
	    return -1;
	}
	
	//
	// Constructor
	//
	
	public JspAppMachine(String instanceId) {
		super();
		this.instanceId = instanceId;
	}

	public JspAppMachine(String instanceId, JspApp app) {
		this(instanceId);
		compile(app);
	}
	
	/**
	 * Compile the final-state machine from app definition:
	 * Call this after making changes to the app definition.
	 *
	 */
	public void compile(JspApp app) {
		initState = currentState = app.getInitialState();
		List ss = app.getStates();
		int n = ss.size();
		fsmStates = new String[n];
		displayActions = new String[n];
		processActions = new String[n];
		inputConditions = new String[n][];
		numTransitions = new int[n];
		cumulativeTransitions = new int[n];
		int tt = 0;
		for(int i=0; i<n; i++) {
			JspAppState s = (JspAppState)ss.get(i);
			fsmStates[i] = s.getName();
			displayActions[i] = s.getDisplay();
			processActions[i] = (s.getProcess() == null ? "" : s.getProcess());
			List ts = s.getTransitions();
			int m = ts.size();
			inputConditions[i] = new String[m];
			for(int j=0; j<m; j++) {
				JspAppTransition t = (JspAppTransition)ts.get(j);
				inputConditions[i][j] = t.getCondition();
			}
			numTransitions[i] = m;
			cumulativeTransitions[i] = tt; // thus must be set before the tt total is incremented
			tt += m;
		}
		transitionTypes = new String[tt];
		nextStates = new String[tt];
		tt=0;
		for(int i=0; i<n; i++) {
			JspAppState s = (JspAppState)ss.get(i);
			List ts = s.getTransitions();
			for(int j=0; j<ts.size(); j++) {
				JspAppTransition t = (JspAppTransition)ts.get(j);
				transitionTypes[tt] = t.getType();
				nextStates[tt] = (t.getNextState()==null ? "" : t.getNextState());
				tt++;
			}
		}
	}
	
	//
	// Accessors
	//
	
	public String getInstanceId() {
		return instanceId;
	}
	
	public String getState() {
		return currentState;
	}
	
	public void setState(String state) {
		currentState = state;
	}
	
	public void setContextValue(String key, Object value) {
		context.put(key, value);
	}
	
	public Object getContextValue(String key) {
		return context.get(key);
	}
	
	public Object removeContextValue(String key) {
		return context.remove(key);
	}

	public Map getContext() {
		return context;
	}
	
	//
	//
	//
	
	public String getProcess() {
		int ndx = getStateIndex(getState());
		if(ndx == -1)
			return "";
		else
			return processActions[ndx];
	}
	
	public String getDisplay() {
		int ndx = getStateIndex(getState());
		if(ndx == -1)
			return displayActions[0];
		else
			return displayActions[ndx];
	}

	/**
	 * Tries to find the condition for the current state. If it fails
	 * it will try with the default condtion. If that fails -1 is returned
	 * @param condition
	 * @return
	 */
	private int getConditionIndex(String condition) {
		int ndx = getStateIndex(getState());
		if(ndx == -1)
			return -1;
		if(condition == null) 
			condition = JspAppNode.JAM_DEFAULT_CONDITION;
		int conditionIndex = getConditionIndex(ndx, condition);
		if(conditionIndex == -1 && !condition.equals(JspAppNode.JAM_DEFAULT_CONDITION)) {
			condition = JspAppNode.JAM_DEFAULT_CONDITION;
			conditionIndex = getConditionIndex(ndx, condition);
		}
		return conditionIndex;
	}
	
	/**
	 * Checks if the given condition is valid for the current state (i.e. is there a
	 * transition for the condition).
	 * @param condition
	 * @return
	 */
	public boolean checkCondition(String condition) {
		return (getConditionIndex(condition)!=-1);
	}
	
	public boolean changeState(String condition) {
		int conditionIndex = getConditionIndex(condition);
		if(conditionIndex == -1)
			return false;
		
		String transitionType = transitionTypes[conditionIndex];
		String nextState = nextStates[conditionIndex];

		if(transitionType.equals(JspAppNode.JAM_CALL_TYPE)) 
			stateStack.push(getState());
		else if(transitionType.equals(JspAppNode.JAM_RETURN_TYPE)) 
			nextState = (String)stateStack.pop();
		setState(nextState);
		
		return true;
	}
	
	/**
	 * Similar to the index-ditto, but this returns the condition selected (or
	 * null if none could be found).
	 * @param condition
	 * @return
	 */
	/*
	private String getCondition(String condition) {
		int ndx = getStateIndex(getState());
		if(ndx == -1)
			return null;
		int conditionIndex = getConditionIndex(ndx, condition);
		if(conditionIndex == -1 && !condition.equals(JspAppNode.JAM_DEFAULT_CONDITION)) {
			condition = JspAppNode.JAM_DEFAULT_CONDITION;
			conditionIndex = getConditionIndex(ndx, condition);
		}
		if(conditionIndex==-1)
			return null;
		else
			return condition;
	}
	*/
	public String dumpFSM(String eol) {
		if(eol==null)
			eol="\n";
		StringBuffer buf = new StringBuffer();
		buf.append("instanceId="+instanceId+eol);
		buf.append("initState="+initState+eol);
		buf.append("fsmStates=");
		for(int i=0; i<fsmStates.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(fsmStates[i]);
		}
		buf.append(eol);
		buf.append("displayActions=");
		for(int i=0; i<displayActions.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(displayActions[i]);
		}
		buf.append(eol);
		buf.append("processActions=");
		for(int i=0; i<processActions.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(processActions[i]);
		}
		buf.append(eol);
		buf.append("inputConditions={");
		for(int i=0; i<inputConditions.length; i++) {
			if(i>0)
				buf.append("................{");
			for(int j=0; j<inputConditions[i].length; j++) {
				if(j>0)
					buf.append(",");
				buf.append(inputConditions[i][j]);
			}
			buf.append("}"+eol);
		}
		buf.append(eol);
		buf.append("numTransitions=");
		for(int i=0; i<numTransitions.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(""+numTransitions[i]);
		}
		buf.append(eol);
		buf.append("cumulativeTransitions=");
		for(int i=0; i<cumulativeTransitions.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(""+cumulativeTransitions[i]);
		}
		buf.append(eol);
		buf.append("transitionTypes=");
		for(int i=0; i<transitionTypes.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(transitionTypes[i]);
		}
		buf.append(eol);
		buf.append("nextStates=");
		for(int i=0; i<nextStates.length; i++) {
			if(i>0)
				buf.append(",");
			buf.append(nextStates[i]);
		}
		buf.append(eol);
		buf.append("stateStack="+stateStack+eol);
		buf.append("currentState="+currentState+eol);
		return buf.toString();
	}

}
