/*
 * Copyright (c) 2006-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 180 $
 * $LastChangedDate: 2007-06-20 11:22:10 +0200 (Wed, 20 Jun 2007) $
 * 
 */

package com.codedroids.oc;

import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import javax.servlet.jsp.PageContext;

import org.opencms.jsp.CmsJspBean;
import org.opencms.main.CmsEvent;
import org.opencms.main.CmsException;
import org.opencms.main.I_CmsEventListener;
import org.opencms.main.OpenCms;

/**
 * Implements the default caching behaviour for contents loaded from the VFS: always online, never offline (though it checks
 * last modification date before actually refreshing the offline cached objects).
 * 
 * <p>Objects that is to be cached must implement the {@link I_OCCacheElement} interface. If it not feasible to create specific
 * subclasses for this purpose, it is possibly to create a wrapper object which then must implement both the {@link I_OCCacheElement}
 * and the {@link I_OCCacheElementWrapper} interfaces. See {@link OCCachedProperties} and {@link OCCachedFile} for examples on
 * subclassing and wrapping, respectively.
 * 
 * <p>It is possible to initialize the cache with a minimum caching period (in seconds) for the offline objects, that otherwise
 * will be checked every time the element is looked up in the cache (typically the check at least includes a VFS operation in 
 * order to determine the last modified date).<br>
 * E.g. if minimum cache period is 600 sec. then the check whether an element is stale or not, will only happen every 10 minutes.
 * 
 * @see I_OCCacheElement
 * @see I_OCCacheElementWrapper
 * @see OCCachedFile
 * @see OCCachedProperties
 *
 */
public class OCCache extends Object implements I_CmsEventListener {

	private static final String KEY = "com.codedroids.oc.OCCache";
	
	/**
	 * Convenience method that will instantiate or reuse a single cache in the application context.
	 * 
	 * <p>The OCCache instance is stored as an attribute in the given application context 
	 * with the key <code>com.codedroids.oc.OCCache</code>. 
	 * 
	 * @param An instantiated CmsJspBean
	 * @return
	 */
	public static OCCache getAppInstance(CmsJspBean cms) {
		return getInstance(cms, PageContext.APPLICATION_SCOPE);
	}
	
	/**
	 * Convenience method that will instantiate or reuse a single cache in the given scope (typically the application or session scope).
	 * 
	 * <p>The OCCache instance is stored as an attribute in the given scope with the key <code>com.codedroids.oc.OCCache</code>.
	 * The minimum cache period for offline resources is set to zero.</p>
	 * 
	 * @param session
	 * @return
	 */
	public static OCCache getInstance(CmsJspBean cms, int scope) {
		PageContext pg = cms.getJspContext();
		OCCache ca = (OCCache)pg.getAttribute(KEY, scope);
		if(ca == null) {
			ca = new OCCache();
			pg.setAttribute(KEY, ca, scope);
		}
		return ca;
	}

	// INSTANCE
	
	private long minimumCachePeriod;
	private Map offlineCache;
	private Map offlineTimestamps;
	private Map onlineCache;
	
	/**
	 * Initialises a new instance setting the minimum offline cache period to one second.
	 * 
	 * <p>The assumption is that such a delay will not be noticed by the user that updates a cached file
	 * in the workplace, and yet it is so long that a page with many elements that looks up a file in the cache
	 * will not be doing a VFS modification date check every time an element is processed).
	 */
	public OCCache() {
		this(1);
	}
	
	/**
	 * Initialises a new instance with the given minimum cache period for the offline part of the cache.
	 * 
	 * @param minimumCachePeriod
	 */
	public OCCache(int minimumCachePeriod) {
		super();
		setOfflineMinCachePeriod(minimumCachePeriod);
		this.offlineCache = new HashMap();
		this.offlineTimestamps = new HashMap();
		this.onlineCache = new HashMap();
		OpenCms.addCmsEventListener(this, new int[] {I_CmsEventListener.EVENT_PUBLISH_PROJECT});
		System.out.println(this.getClass().getName()+" registered as listener.");
	}
	
	/**
	 * <p>The minimum cache period for offline elements. 
	 * When an object is stored in the cache it is timestamped. The minimum cache period ensures that in that period 
	 * of time after the objects timestamp, no attempt will be done to evaluate whether or not the object is stale.</p>
	 * <p>Since the stale check in most cases includes a check of the last modified attribute of at least one file from
	 * the VFS, this can save some time in case of a workplace tool used in production but still used from within an offline
	 * project.</p>
	 * <p>Note, this can be set at any time and will influence already cached values immediatly. This can be useful for
	 * updating an app which already has a long minimum period but that needs to refresh its cached elements asap.</p>
	 * <p>The online part of the cache is not influenced by this setting</p>
	 * 
	 * @param minimumCachePeriod
	 */
	public synchronized void setOfflineMinCachePeriod(int minimumCachePeriod) {
		this.minimumCachePeriod = minimumCachePeriod * 1000L;
	}
	
	/**
	 * Convenience that returns the value from the I_OCCacheElement wrapper rather than the element itself
	 * if it is a wrapped object.
	 * 
	 * @param cms
	 * @param key
	 * @param elementClass
	 * @param context
	 * @return
	 * @throws CmsException
	 */
	public Object cacheObject(CmsJspBean cms, String key, Class elementClass, Object context) throws CmsException {
		I_OCCacheElement el = cacheElement(cms, key, elementClass, context);
		if(el == null)
			return null;
		else if(el instanceof I_OCCacheElementWrapper)
			return ((I_OCCacheElementWrapper)el).getCacheValue();
		else
			return el;		
	}
	
	/**
	 * Convenience that returns the value from the I_OCCacheElementWrapper rather than the element itself
	 * in case object implements that interface (otherwise the cached object is returned as expected)
	 * 
	 * @param cms
	 * @param key
	 * @param elementClass
	 * @param context
	 * @return
	 * @throws CmsException
	 */
	public Object getCachedObject(CmsJspBean cms, String key, Class elementClass, Object context) throws CmsException {
		I_OCCacheElement el = getCachedElement(cms, key, elementClass, context);
		if(el == null)
			return null;
		else if(el instanceof I_OCCacheElementWrapper)
			return ((I_OCCacheElementWrapper)el).getCacheValue();
		else
			return el;		
	}
	
	/**
	 * Create a new element in the cache (possibly overwriting an existing one if the given key is already in use)
	 * 
	 * @param cms
	 * @param key
	 * @param elementClass
	 * @param context
	 * @throws CmsException
	 */
	public synchronized I_OCCacheElement cacheElement(CmsJspBean cms, String key, Class elementClass, Object context) throws CmsException {
		try {
			I_OCCacheElement el = (I_OCCacheElement)elementClass.newInstance();
			boolean online = cms.getRequestContext().currentProject().isOnlineProject();
			if(el.init(cms, online, key, context)) {
				if(online) {
					onlineCache.put(key, el);
				} else {
					offlineCache.put(key, el);
					offlineTimestamps.put(key, new Long(System.currentTimeMillis() + minimumCachePeriod));
				}
				return el;
			} else {
				return null;
			}
		} catch (InstantiationException e) {
			throw new RuntimeException("Element class = " + elementClass.getName(), e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Element class = " + elementClass.getName(), e);
		}
	}
	
	/**
	 * Looks up the element, and if not found, it will be loaded/cached or refreshed as needed.
	 * @param cms
	 * @param path
	 * @param elementClass
	 * @param context
	 * @return
	 * @throws CmsException
	 */
	public synchronized I_OCCacheElement getCachedElement(CmsJspBean cms, String key, Class elementClass, Object context) throws CmsException {
		I_OCCacheElement el = null;
		boolean online = cms.getRequestContext().currentProject().isOnlineProject();
		if(online) {
			el = (I_OCCacheElement)onlineCache.get(key);
			if(el == null) {
				el = cacheElement(cms, key, elementClass, context);
			}
			else if(el.isStale(cms)) {
				if(!el.refresh(cms)) {
					onlineCache.remove(key);
					el = null;
				}
			}
		} else {
			el = (I_OCCacheElement)offlineCache.get(key);
			if(el == null) {
				el = cacheElement(cms, key, elementClass, context);
			}
			else if(((Long)offlineTimestamps.get(key)).longValue() < System.currentTimeMillis()) {
				if(el.isStale(cms)) {
					if(!el.refresh(cms)) {
						offlineCache.remove(key);
						offlineTimestamps.remove(key);
						el = null;
					} else {
						// remember to refresh the timestamp
						offlineTimestamps.put(key, new Long(System.currentTimeMillis() + minimumCachePeriod));
					}
				}
			}
		}
		return el;
	}
	
	/**
	 * This only looks up and return an already cached element, will not load or refresh the element.
	 * You should check that the returned element is not stale.
	 * @param cms
	 * @param key
	 * @return 
	 * @throws CmsException
	 */
	public synchronized I_OCCacheElement getCachedElement(CmsJspBean cms, String key) throws CmsException {
		I_OCCacheElement el = null;
		boolean online = cms.getRequestContext().currentProject().isOnlineProject();
		if(online) {
			el = (I_OCCacheElement)onlineCache.get(key);
		} else {
			el = (I_OCCacheElement)offlineCache.get(key);
		}
		return el;
	}

	
	public synchronized void clearOnlineCache() {
		onlineCache.clear();
	}
	public synchronized void clearOfflineCache() {
		offlineCache.clear();
	}
	public synchronized void clearCaches() {
		clearOnlineCache();
		clearOfflineCache();
	}
	
	// could be more discriminate since we know what files are being published...
	public void cmsEvent(CmsEvent event) {
		switch (event.getType()) {
			case I_CmsEventListener.EVENT_PUBLISH_PROJECT:
				clearCaches();
				break;
		}
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(this.getClass().getName() + " caches " + onlineCache.size() + " online and " + offlineCache.size() + " offline objects (minimum cache period = " + minimumCachePeriod + " sec.)\n");
		if(onlineCache.size() > 0) {
			buf.append("Online:\n");
			Iterator it = onlineCache.keySet().iterator();
			while(it.hasNext()) {
				buf.append("  " + (String)it.next() + "\n");
			}
		}
		if(offlineCache.size() > 0) {
			buf.append("Offline:\n");
			Iterator it = offlineCache.keySet().iterator();
			while(it.hasNext()) {
				buf.append("  " + (String)it.next() + "\n");
			}
		}
		return buf.toString();
	}

}
